////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2012 Kentaro Sekimoto  All rights reserved.
////////////////////////////////////////////////////////////////////////////

/*
 * FM3_EMAC_lwip.cpp
 *
 *  Created on: 2012/02/27
 *      Author: Kentaro Sekimoto
 */
#include <tinyhal.h>
#include "net_decl_lwip.h"
#include "..\FM3.h"
#include "FM3_EMAC_LWIP.h"
#include "FM3_EMAC_LWIP_driver.h"

extern void lwip_interrupt_continuation(void);
extern struct netif g_FM3_EMAC_NetIF;
#ifdef INTERRUPT_TASK_CONTINUATION
extern HAL_CONTINUATION    InterruptTaskContinuation;
#else
extern HAL_COMPLETION    InterruptTaskCompletion;
#endif

// ========== ftHgl  Jn ==========
#define ETH_BUF_MIN
// DMA Register 0 (BMR)
#define DEF_DMA_PBL     16      // Programmable Burst Length
#define DEF_DMA_DSL     0       // Ring Mode
#define DEF_DMA_PR      1       // 1 -> Rx:Tx=2:1
// MfBXNv^
#ifdef ETH_BUF_MIN
#define DEF_TXDESC_NUM  3       // MfBXNv^̐
#define DEF_RXDESC_NUM  4       // MfBXNv^̐
#define DEF_TX_BUF_NUM  DEF_TXDESC_NUM      // Mobt@̐
#define DEF_TX_BUF_SIZE 1536                // Mobt@̃TCY
#define DEF_RX_BUF_SIZE 1536                // Mobt@̐
#define DEF_RX_BUF_NUM  DEF_RXDESC_NUM      // Mobt@̃TCY
#else
#define DEF_TXDESC_NUM  4       // MfBXNv^̐
#define DEF_RXDESC_NUM  4       // MfBXNv^̐
#define DEF_TX_BUF_NUM  (DEF_TXDESC_NUM*2)  // Mobt@̐
#define DEF_TX_BUF_SIZE 1536                // Mobt@̃TCY
#define DEF_RX_BUF_SIZE 1536                // Mobt@̐
#define DEF_RX_BUF_NUM  (DEF_RXDESC_NUM*2)  // Mobt@̃TCY
#endif
#define MFFR_PROMISCUOUS                    // PROMISCUOUS[h
// ========== ftHgl  I ==========

// ========== PHY Jn ==========
/* LAN8187C PHY Registers */
#define PHY_REG_BMCR    0x00    // Basic Mode Control Register
#define PHY_REG_BMSR    0x01    // Basic Mode Status Register
#define PHY_REG_IDR1    0x02    // PHY Identifier 1
#define PHY_REG_IDR2    0x03    // PHY Identifier 2
#define PHY_REG_STS     0x10    // Status Register
#define PHY_REG_19      19      // Extended Control Register 19
#define PHY_REG_LEDC    20
#define PHY_REG_INTC    22

#define BMCR_RESET      0x8000
#define BMSR_AUTO_DONE  0x0020

//#define PHY_AUTO_NEG  0x1000    // Select Auto Negotiation
#define PHY_AUTO_NEG    0x1200      // Select Auto Negotiation

#define STS_LINK_ON     0x1
#define STS_10_MBIT     0x2
#define STS_FULL_DUP    0x4

#define MII_WR_TOUT     0x00200000  // MII Write timeout count
#define MII_RD_TOUT     0x00200000  // MII Read timeout count

#define LAN8187_ADDR    0x06        // PHY device address for Wakamatsu FM3 LAN board
#define ICS1894_ADDR    0x10        // PHY device address for Wakamatsu NXP LAN board
//#define LAN8187_ID    0x20005C90  // PHY Identifier of LAN8187
//#define ICS1894_ID    0x0015f450  // PHY Identifier of ICS1894
#define PHY_ADDR        ICS1894_ADDR
#define DEF_PHY_RESET_RETRY     5
#define DEF_PHY_STATUS_RETRY    100000
// ========== PHY I ==========

// ========== ϐ Jn ==========
// FM3WX^
FM3_USBETHERNETCLK_TypeDef *ethclk = FM3_USBETHERNETCLK;
FM3_ETHERNET_MAC_TypeDef *ethmac = FM3_ETHERNET_MAC0;
FM3_ETHERNET_CONTROL_TypeDef *ethctrl = FM3_ETHERNET_CONTROL;

// Mobt@
char TxBuf[DEF_TX_BUF_NUM][DEF_TX_BUF_SIZE] __attribute__((aligned(4)));
char RxBuf[DEF_RX_BUF_NUM][DEF_RX_BUF_SIZE] __attribute__((aligned(4)));

// MfBXNv^
EMAC_DMA_TXDESC txdesc[DEF_TXDESC_NUM] __attribute__((aligned(4)));
EMAC_DMA_RXDESC rxdesc[DEF_RXDESC_NUM] __attribute__((aligned(4)));

char tmpbuf[DEF_TX_BUF_SIZE] __attribute__((aligned(4)));

unsigned int txdesc_id = 0;    // MfBXNv^݂̌̃CfbNX

// MACAhX
char MACAddr[6];
char Broadcast[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

// ========== ϐ I==========

#if 0
static unsigned short htons(unsigned short n)
{
    return ((n & 0xff) << 8) | ((n & 0xff00) >> 8);
}

static unsigned short ntohs(unsigned short n)
{
    return ((n & 0xff) << 8) | ((n & 0xff00) >> 8);
}
#endif

static void wait(volatile unsigned int count)
{
    while (count-- > 0) ;
}

int PHY_Write(unsigned int phy_addr, unsigned int phy_register, unsigned int value, unsigned int timeout)
{
    while ((ethmac->GAR & GMAC_GAR_GB) != 0) {
        if (timeout-- == 0)
            return FALSE;
    }
    ethmac->GDR = (unsigned short)value;
    ethmac->GAR &= ~((0x1f << GMAC_GAR_B_PA) | (0x1f << GMAC_GAR_B_GR));
    ethmac->GAR |= ((phy_addr << GMAC_GAR_B_PA) | (phy_register << GMAC_GAR_B_GR));
    ethmac->GAR |=  ((1 << GMAC_GAR_B_GW) |(1 << GMAC_GAR_B_GB));
    return TRUE;
}

int PHY_Read(unsigned int phy_addr, unsigned int phy_register, unsigned int *value, unsigned int timeout)
{
    unsigned int count = timeout;
    while ((ethmac->GAR & GMAC_GAR_GB) != 0) {
        if (count-- == 0)
            return FALSE;
    }
    ethmac->GAR &= ~((0x1f << GMAC_GAR_B_PA) | (0x1f << GMAC_GAR_B_GR));
    ethmac->GAR |= ((phy_addr << GMAC_GAR_B_PA) | (phy_register << GMAC_GAR_B_GR));
    ethmac->GAR &=  ~(1 << GMAC_GAR_B_GW);
    ethmac->GAR |= (1 << GMAC_GAR_B_GB);
    count = timeout;
    while ((ethmac->GAR & GMAC_GAR_GB) != 0) {
        if (count-- == 0)
            return FALSE;
    }
    *value = (unsigned int)ethmac->GDR;
    return TRUE;
}

int PHY_Reset(void)
{
    unsigned int value;
    int flag = FALSE;
    unsigned int reset_retry = DEF_PHY_RESET_RETRY;
    while (reset_retry-- > 0) {
        if (PHY_Write(PHY_ADDR, PHY_REG_BMCR, BMCR_RESET, MII_WR_TOUT) == TRUE) {
            unsigned int status_retry = DEF_PHY_STATUS_RETRY;
            while (status_retry-- > 0) {
                if (PHY_Read(PHY_ADDR, PHY_REG_BMCR, &value, MII_RD_TOUT) == TRUE) {
                    if ((value & BMCR_RESET) == 0) {
                        flag = TRUE;
                        break;
                    }
                }
            }
            if (flag == TRUE)
                break;
        }
    }
    return flag;
}

int PHY_AutoNegotiate(void)
{
    unsigned int value;
    int flag = FALSE;
    unsigned int reset_retry = DEF_PHY_RESET_RETRY;
    while (reset_retry-- > 0) {
        if (PHY_Write(PHY_ADDR, PHY_REG_BMCR, PHY_AUTO_NEG, MII_WR_TOUT) == TRUE) {
            //PHY_Write(PHY_ADDR, 4, 0x1E1, MII_WR_TOUT);
            //PHY_Write(PHY_ADDR, PHY_REG_BMCR, 0x1200, MII_WR_TOUT);
            unsigned int status_retry = DEF_PHY_STATUS_RETRY;
            while (status_retry-- > 0) {
                if (PHY_Read(PHY_ADDR, PHY_REG_BMSR, &value, MII_RD_TOUT) == TRUE) {
                    if (value & BMSR_AUTO_DONE) {
                        flag = TRUE;
                        break;
                    }
                }
            }
            if (flag == TRUE)
                break;
        }
    }
    if (flag)
        hal_printf("PHY: AutoNegotiate OK\r\n");
    else
        hal_printf("PHY: AutoNegotiate NG\r\n");
    return flag;
}

int PHY_Init(void)
{
    int i;
    unsigned int value;
    unsigned int id1, id2;
    int flag = FALSE;

    if (PHY_Reset() == TRUE) {
        hal_printf("PHY: Reset OK\r\n");
    } else {
        hal_printf("PHY: Reset NG\r\n");
        return flag;
    }
    if (PHY_Read(PHY_ADDR, PHY_REG_IDR1, &id1, MII_RD_TOUT) == TRUE) {
        hal_printf("PHY: ReadID OK %04x\r\n", id1);
    } else {
        hal_printf("PHY: ReadID NG\r\n");
    }
    if (PHY_Read(PHY_ADDR, PHY_REG_IDR2, &id2, MII_RD_TOUT) == TRUE) {
        hal_printf("PHY: ReadID OK %04x\r\n", id2);
    } else {
        hal_printf("PHY: ReadID NG\r\n");
    }
#if 1
    if (PHY_Write(PHY_ADDR, PHY_REG_19, 0x0201, MII_WR_TOUT) == TRUE) {
        hal_printf("PHY: MDI_MODE OK\r\n");
    } else {
        hal_printf("PHY: MDI_MODE NG\r\n");
    }
#endif
#if 1
    /* LED0/LED1 _p^[ݒ(LED0:NAbv/LED1:ANeBu) */
    if (PHY_Write(PHY_ADDR, PHY_REG_LEDC, 0x0008, MII_WR_TOUT) == TRUE) {
        hal_printf("PHY: LED initialize OK\r\n");
    } else {
        hal_printf("PHY: LED initialize NG\r\n");
    }
#endif
#if 0
    /* PHY荞ݏo͐ݒ(荞ݏo͗L/NAbv&N_E荞) */
    if (PHY_Write(PHY_ADDR, PHY_REG_INTC, 0x8005, MII_WR_TOUT) == TRUE) {
        hal_printf("PHY: IntOut initialize OK\r\n");
    } else {
        hal_printf("PHY: IntOut initialize NG\r\n");
    }
#endif
#if 0
    // Bias  90% 0x0521
    // Bias 100% 0x0641
    // Bias 110% 0x0761
    if (PHY_Write(PHY_ADDR, 25, 0x0761, MII_WR_TOUT) == TRUE) {
        hal_printf("PHY: 110%% Bias Setting OK\r\n");
    } else {
        hal_printf("PHY: 110%% Bias Setting NG\r\n");
    }
#endif
#ifdef DEBUG_DUMP
    PHY_Register_Dump();
#endif
    flag = TRUE;
    return flag;
}

int PHY_LinkSpeed(void)
{
    unsigned int value;
    int full_duplex, mbit_100;

    hal_printf("Waiting for auto negotiation...\r\n");
    if (PHY_AutoNegotiate() != TRUE) {
        mbit_100 = TRUE;
        full_duplex = TRUE;
#if 1
        PHY_Write(PHY_ADDR, PHY_REG_BMCR, 0x3100, MII_WR_TOUT);
        ethmac->MCR |= (GMAC_MCR_FES | GMAC_MCR_DM);
        hal_printf("LinkSpeed: Full Duplex\r\n");
        hal_printf("LinkSpeed: 100Mbs\r\n");
#endif
    } else {
        if (PHY_Read(PHY_ADDR, PHY_REG_BMSR, &value, MII_RD_TOUT) == FALSE){
            return FALSE;
        }
        if (value & STS_FULL_DUP) {
            hal_printf("LinkSpeed: Full Duplex\r\n");
            full_duplex = TRUE;
        } else {
            hal_printf("LinkSpeed: Half Duplex\r\n");
            full_duplex = FALSE;
        }
        if (value & STS_10_MBIT) {
            hal_printf("LinkSpeed: 10Mbs\r\n");
            mbit_100 = FALSE;
        } else {
            hal_printf("LinkSpeed: 100Mbs\r\n");
            mbit_100 = TRUE;
            PHY_Write(PHY_ADDR, PHY_REG_BMCR, 0x2100, MII_WR_TOUT);
            ethmac->MCR |= (GMAC_MCR_FES | GMAC_MCR_DM);
        }
    }
    return TRUE;
}

// MfBXNv^̏
void TXDESC_Init(EMAC_DMA_TXDESC *txdesc, unsigned int b1ap, unsigned int b2ap)
{
    txdesc->TDES0 = 0;
    txdesc->TDES1_f.TBS1 = 0;
    txdesc->TDES1_f.TBS2 = 0;
    txdesc->TDES2 = b1ap;
    txdesc->TDES3 = b2ap;
    txdesc_id = 0;
}

// MfBXNv^̏
void RXDESC_Init(EMAC_DMA_RXDESC *rxdesc, unsigned int b1ap, unsigned int b2ap)
{
    rxdesc->RDES0 = 0;
    rxdesc->RDES1_f.RBS1 = DEF_RX_BUF_SIZE;
#ifdef ETH_BUF_MIN
    rxdesc->RDES1_f.RBS2 = 0;
#else
    rxdesc->RDES1_f.RBS2 = DEF_RX_BUF_SIZE;
#endif
    rxdesc->RDES2 = b1ap;
    rxdesc->RDES3 = b2ap;
}

// MfBXNv^̏
void DESC_Init(void)
{
    unsigned int i;
#ifdef ETH_BUF_MIN
    for (i=0; i<DEF_TXDESC_NUM; i++) {
        TXDESC_Init(&txdesc[i], (unsigned int)&TxBuf[i], (unsigned int)0);
        if (i == (DEF_TXDESC_NUM-1))
            txdesc[i].TDES0_f.TER = 1;
    }
    for (i=0; i<DEF_RXDESC_NUM; i++) {
        RXDESC_Init(&rxdesc[i], (unsigned int)&RxBuf[i], (unsigned int)0);
        rxdesc[i].RDES0_f.OWN = 1;
        if (i == (DEF_RXDESC_NUM-1))
            rxdesc[i].RDES1_f.RER = 1;
    }
#else
    for (i=0; i<DEF_TXDESC_NUM; i++) {
        TXDESC_Init(&txdesc[i], (unsigned int)&TxBuf[i*2], (unsigned int)&TxBuf[i*2+1]);
        if (i == (DEF_TXDESC_NUM-1))
            txdesc[i].TDES0_f.TER = 1;
    }
    for (i=0; i<DEF_RXDESC_NUM; i++) {
        RXDESC_Init(&rxdesc[i], (unsigned int)&RxBuf[i*2], (unsigned int)&RxBuf[i*2+1]);
        rxdesc[i].RDES0_f.OWN = 1;
        if (i == (DEF_RXDESC_NUM-1))
            rxdesc[i].RDES1_f.RER = 1;
    }
#endif
    ethmac->RDLAR = (unsigned int)&rxdesc[0];
    ethmac->TDLAR = (unsigned int)&txdesc[0];
}

// Ethernet DMȀ
// Refer to Programming Guide P.170
void DMA_Init(void)
{
    ethmac->BMR = DMA_BMR_SWR;                      // soft reset
    while ((ethmac->BMR & DMA_BMR_SWR) == 1);       // wait for reset completion (== 0)
    while ((ethmac->AHBSR & DMA_AHBSR_AHBS) == 1);  // wait for AHB master transaction completion
    ethmac->MCR |= GMAC_MCR_PS;                     // select MII/RMII interface (MUST)
    // DMA Bus Mode Register
    // Recommend: Mixed Burst=1 , AAL=0, FB = 0, Fixed burst or undefined burst.
#if 0
    ethmac->BMR &= ~(DMA_BMR_FB | DMA_BMR_AAL);
    ethmac->BMR |= (DMA_BMR_MB |                    // Mixed Burst = 1
            (DEF_DMA_PBL << DMA_BMR_B_PBL) |        // Programmable Burst Length
            (DEF_DMA_DSL << DMA_BMR_B_DSL) |        // Descriptor Skip Length
            DMA_BMR_DA |                            // DMA arbitration scheme
            (DEF_DMA_PR << DMA_BMR_B_PR));          // Priority Ratio
#else
    ethmac->BMR = 0x04025002;
#endif
    DESC_Init();
    // DMA Operation Mode Register
    // RSF (Receive Store and Forward)
    // TSF (Transmit Store Forward)
    // FUF (Forward Undersized Good Frames) 64oCg̓]
    ethmac->OMR = (DMA_OMR_RSF | DMA_OMR_TSF | DMA_OMR_FUF);
    ethmac->SR |= DMA_SR_NIS;                       // clear normal interrupt status
    // NIE (Normal Interrupt Summary Enable)
    // ERE (Early Receive Interrupt Enable)
    // RIE (Receive Interrupt Enable)
    // TIE (Transmit Interrupt)
    ethmac->IER = (DMA_IER_NIE |DMA_IER_TIE | DMA_IER_RIE | DMA_IER_ERE);
    while ((ethmac->AHBSR & DMA_AHBSR_AHBS) == 1);  // wait for AHB master transaction completion
    // SR DMAMRunning
    // ST DMAMRunning
    ethmac->OMR |= (DMA_OMR_SR | DMA_OMR_ST);
    //ethmac->OMR |= DMA_OMR_SR;
}

// GMAC̏
// GMAC : Ethernet Media Access Controller
void GMAC_Init(unsigned char *pmac)
{
    ethmac->GAR |= GMAC_GAR_CR;             // CR:0001 -> SYS_CLK 100-150MHz SYS_CLK/62
    PHY_Init();                             // PHY Initialize
    PHY_LinkSpeed();                        // Set LinkSpeed
    memcpy((void *)MACAddr, (void *)pmac, 6);
    ethmac->MAR0H = (unsigned int)(*((unsigned short *)pmac));
    ethmac->MAR0L = (unsigned int)(*((unsigned int *)(pmac+2)));
#ifdef MFFR_PROMISCUOUS
    ethmac->MFFR = (GMAC_MFFR_PR | GMAC_MFFR_RA);
#else
    ethmac->MAR1H = (unsigned int)(*((unsigned short *)pmac)) | 0x80000000;
    ethmac->MAR1L = (unsigned int)(*((unsigned int *)(pmac+2)));
    ethmac->MFFR &= ~(GMAC_MFFR_DB);
    ethmac->MFFR |= (GMAC_MFFR_HUC | GMAC_MFFR_RA);
#endif
    ethmac->MCR |= GMAC_MCR_TE | GMAC_MCR_RE;   // enable transmit and receive
}

// Ethernet Control WX^
void ETHCTRL_Init(void)
{
    ethctrl->ETH_CLKG = (3 << ETH_CLKG_MACEN);      // start EMAC clock
#ifdef ETH_CH0
    ethctrl->ETH_MODE = (1 << ETH_MODE_IFMODE) |    // set RMII mode and reset EMAC
            (1 << ETH_MODE_RST0);
    wait(500000);
    ethctrl->ETH_MODE = (1 << ETH_MODE_IFMODE) |    // set RMII mode and start EMAC
            (0 << ETH_MODE_RST0);
#endif
#ifdef ETH_CH1
    ethctrl->ETH_MODE = (1 << ETH_MODE_IFMODE) |    // set RMII mode and reset EMAC
            (1 << ETH_MODE_RST1);
    wait(500000);
    ethctrl->ETH_MODE = (1 << ETH_MODE_IFMODE) |    // set RMII mode and start EMAC
            (0 << ETH_MODE_RST1);
#endif
}

// p\ȑMfBXNv^̎擾
EMAC_DMA_TXDESC *TXDESC_Available(void)
{
    int i = DEF_TXDESC_NUM;
    EMAC_DMA_TXDESC *ptxdesc;
    while (i-- > 0) {
        ptxdesc = (EMAC_DMA_TXDESC *)&txdesc[txdesc_id];
        txdesc_id++;
        if (txdesc_id == DEF_TXDESC_NUM)
            txdesc_id = 0;
        if (ptxdesc->TDES0_f.OWN == 0)
            return ptxdesc;
    }
#if 0
    for (i=0; i<DEF_TXDESC_NUM; i++) {
        if (txdesc[i].TDES0_f.OWN == 0)
            return (EMAC_DMA_TXDESC *)&txdesc[i];
    }
#endif
    return (EMAC_DMA_TXDESC *)NULL;
}

// Mf[^MfBXNv^̎擾
EMAC_DMA_RXDESC *RXDESC_Received(void)
{
    int i;
    for (i=0; i<DEF_RXDESC_NUM; i++) {
        if (rxdesc[i].RDES0_f.OWN == 0) {
            return (EMAC_DMA_RXDESC *)&rxdesc[i];
        }
    }
    return (EMAC_DMA_RXDESC *)NULL;
}

// Mf[^𑗐MfBXNv^ɃRs[AMB(M҂ɂ)
void Packet_Send(unsigned char *p, unsigned int size)
{
    //lcd_printf("PSend: %04X\r\n", size);
    EMAC_DMA_TXDESC *ptxdesc;
    while ((ptxdesc = TXDESC_Available()) == NULL);
    unsigned char *dst = (unsigned char *)ptxdesc->TDES2;
    memcpy(dst, p, size);
    //ptxdesc->TDES0_f.CIC = 3;         // `FbNT̎ǉ
    ptxdesc->TDES0_f.CIC = 0;           // `FbNT̎ǉȂ
    //ptxdesc->TDES0_f.IC = 1;          //@݂̃t[MAM荞
    ptxdesc->TDES0_f.FS = 1;            // obt@t[̍ŏ̕܂
    ptxdesc->TDES0_f.LS = 1;            // obt@t[̍Ō̕܂
    ptxdesc->TDES1_f.TBS1 = size;       // MTCYݒ
    ptxdesc->TDES0_f.OWN = 1;           // MĊĴ߁AfBXNv^DMALƂ
    ethmac->OMR_f.ST = 1;               // DMARunXe[g
    ethmac->TPDR = (unsigned int)0;     // MĊĴ߁AfBXNv^̃|[Ov
}

//int strncmp(void *p1, void *p2, int len);

// Ethernet݃nh
void ETH_IRQHandler(FM3_ETHERNET_MAC_TypeDef *pethmac)
{
    int i;
    EMAC_DMA_TXDESC *ptxdesc = 0;
    EMAC_DMA_RXDESC    *prxdesc = 0;
    unsigned int gmac_isr = pethmac->ISR;
    unsigned int dma_sr = pethmac->SR;
    if ((dma_sr & DMA_SR_RI) != 0) {
        if ((prxdesc = RXDESC_Received()) != NULL) {
            struct eth_packet *ethp = (struct eth_packet *)prxdesc->RDES2;
#ifdef ONLY_DIRECT_AND_BROADCAST
            if ((hal_strncmp_s((const char*)ethp, (const char*)&MACAddr[0], 6) == 0) ||
                (hal_strncmp_s((const char*)ethp, (const char*)&Broadcast[0], 6) == 0))
            {
                FM3_EMAC_LWIP_recv(&g_FM3_EMAC_NetIF);
            } else {
                prxdesc->RDES0_f.OWN = 1;
            }
#else
            FM3_EMAC_LWIP_recv(&g_FM3_EMAC_NetIF);
#endif
            pethmac->SR |= DMA_SR_RI;
        }
    } else
        if ((dma_sr & DMA_SR_TU) != 0) {
            // TU (Transmit Buffer Unavailable)̏
            // MAAđMۂɗp
            // TODO
            pethmac->SR |= (DMA_SR_TU);     // TCR_SS_WCrbg
            hal_printf("TU\n");
        } else if ((dma_sr & DMA_SR_TI) != 0) {
            // TI TI (Transmit Interrupt)̏
            // MAAđMۂɗp
            // TODO
            //ethmac->SR |= (DMA_SR_TI);    // TIR_SS_WCrbg
            hal_printf("TI\n");
        } else if ((dma_sr & DMA_SR_ERI) != 0) {
            // ERI (Early Receive Interrupt))
            // DMA pPbg̍ŏ̃f[^obt@[UƂBNA
            //hal_printf("ISR:%08x DSR:%08x\r\n", gmac_isr, dma_sr);
        } else {
            // fobOp\
            hal_printf("ISR:%08x DSR:%08x\r\n", gmac_isr, dma_sr);
        }
}

#ifdef ETH_CH0
void ETH0_IRQHandler(void* param) __attribute__((interrupt));
void ETH0_IRQHandler(void* param)
{
    ETH_IRQHandler(FM3_ETHERNET_MAC0);
}
#endif

#ifdef ETH_CH1
void ETH1_IRQHandler(void* param) __attribute__((interrupt));
void ETH1_IRQHandler(void* param)
{
    ETH_IRQHandler(FM3_ETHERNET_MAC1);
}
#endif

// Pin  CH  Pin name
// PC0: 1   E_RXDV1
// PC1: 1   E_RX11
// PC2: 1   E_RX10
// PC3: 0   E_RX01
// PC4: 0   E_RX00
// PC5: 0   E_RXDV0
// PC6: 0   E_MDIO0
// PC7: 0   E_MDC0
// PC8: 01  E_RXCK0_REFCK
// PC9:
// PCA:
// PCB:
// PCC: 1   E_MDIO1
// PCD: 1   E_MDC1
// PCE: 1   E_TXEN1
// PCF: 1   E_TX11
// PD0: 1   E_TX10
// PD1: 0   E_TX01
// PD2: 0   E_TX00
// PD3: 0   E_TXEN0

//#define PFRC_ETH0 0x01F8
//#define PFRD_ETH0 0x000E
#define PFRC_ETH    0xF1FF
#define PFRD_ETH    0x000F

#define B_E_TD0E    18
#define B_E_TD1E    19
#define B_E_TE0E    20
#define B_E_TE1E    21
#define B_E_MC0E    22
#define B_E_MC1B    23
#define B_E_MD0B    24
#define B_E_MD1B    25
#define B_E_CKE     26
#define B_E_PSE     27
#define B_E_SPLC    28

// Ethernet֘AGPIOs̏ъݐݒ
void ETHHW_Init(void)
{
    FM3_GPIO->PFR6 &= ~0x0004;
    FM3_GPIO->DDR6 &= ~0x0004;
    FM3_GPIO->PFRC = PFRC_ETH;      // GPIO configuration for CH0 and CH1
    FM3_GPIO->PFRD = PFRD_ETH;      // GPIO configuration for CH0 and CH1
    FM3_GPIO->EPFR14 = (3 << B_E_SPLC) |
                    // (1 << B_E_PSE) |
            (1 << B_E_CKE) |
            (1 << B_E_MD1B) |
            (1 << B_E_MD0B) |
            (1 << B_E_MC1B) |
            (1 << B_E_MC0E) |
            (1 << B_E_TE1E) |
            (1 << B_E_TE0E) |
            (1 << B_E_TD1E) |
            (1 << B_E_TD0E);
}

// ========== fobO֘A ==========

// MfBXNv^̃_v
void TXDESC_Dump(void)
{
    int i;
    for (i=0; i<DEF_TXDESC_NUM; i++) {
        hal_printf("TD[%d]%08x TDES0 %08x\r\n", i, &txdesc[i], txdesc[i].TDES0);
#if 0
        hal_printf("TD[%d]%08x TDES1 %08x\r\n", i, &txdesc[i], txdesc[i].TDES1);
        hal_printf("TD[%d]%08x TDES2 %08x\r\n", i, &txdesc[i], txdesc[i].TDES2);
        hal_printf("TD[%d]%08x TDES3 %08x\r\n", i, &txdesc[i], txdesc[i].TDES3);
#endif
    }
}

// MfBXNv^̃_v
void RXDESC_Dump(void)
{
    int i;
    for (i=0; i<DEF_RXDESC_NUM; i++) {
        hal_printf("RD[%d]%08x RDES0 %08x\r\n", i, &rxdesc[i], rxdesc[i].RDES0);
#if 0
        hal_printf("RD[%d]%08x RDES1 %08x\r\n", i, &rxdesc[i], rxdesc[i].RDES1);
        hal_printf("RD[%d]%08x RDES2 %08x\r\n", i, &rxdesc[i], rxdesc[i].RDES2);
        hal_printf("RD[%d]%08x RDES3 %08x\r\n", i, &rxdesc[i], rxdesc[i].RDES3);
#endif
    }
}

// ݂̃fBXNv^̃_v
void CurrentDESC_Dump(void)
{
    //  EMAC_DMA_TXDESC *ptxdesc = (EMAC_DMA_TXDESC *)ethmac->CHTDR;
    //  EMAC_DMA_RXDESC *prxdesc = (EMAC_DMA_TXDESC *)ethmac->CHRDR;
}

// vDMAWX^̃_v
void DMA_Register_Dump(void)
{
    //hal_printf("GMAC 0 (MCR)  %08x\r\n", ethmac->MCR);
    //hal_printf("GMAC 1 (MFFR) %08x\r\n", ethmac->MFFR);
    //hal_printf("DMA 0  (BMR)  %08x\r\n", ethmac->BMR);
    hal_printf("DMA 5  (SR)   %08x\r\n", ethmac->SR);
    hal_printf("DMA 6  (OMR)  %08x\r\n", ethmac->OMR);
    hal_printf("DMA 7  (IER)  %08x\r\n", ethmac->IER);
    hal_printf("DMA 14 (ISR)  %08x\r\n", ethmac->ISR);
    hal_printf("DMA 18 (CHDR) %08x\r\n", ethmac->CHTDR);
    hal_printf("txdesc        %08x\r\n", &txdesc[txdesc_id]);
}

unsigned int phy_register_mask = 0b00000011111111110000000111111111;

// vPHYWX^̃_v
void PHY_Register_Dump(void)
{
    unsigned int value;
    unsigned int bit_mask = phy_register_mask;
    int i;
    for (i = 0; i < 31; i++) {
        PHY_Read(PHY_ADDR, (unsigned int)i, &value, MII_RD_TOUT);
        if (bit_mask & 0x1)
            hal_printf("Reg %02d: %04x\r\n", i, value);
        bit_mask >>= 1;
    }
}

// ====================

BOOL FM3_EMAC_LWIP_Init(struct netif *pNetIf)
{
    ETHHW_Init();
    DESC_Init();
    ETHCTRL_Init();
    DMA_Init();
    GMAC_Init((unsigned char *)&pNetIf->hwaddr);
    return true;
//    return (BOOL)PHY_Init();
}

BOOL FM3_EMAC_LWIP_open(struct netif *pNetIf)
{
    return FM3_EMAC_LWIP_Init(pNetIf);
}

void FM3_EMAC_LWIP_close(struct netif *pNetIf)
{
    // TODO
}

err_t FM3_EMAC_LWIP_xmit(struct netif *pNetIf, struct pbuf *pPBuf)
{
    UINT32 bytes_out, remains;
    UINT8 *dst;

    GLOBAL_LOCK(irq);
    if (!pNetIf || !pPBuf)
        return ERR_ARG;
    bytes_out = (UINT32)pPBuf->tot_len;
    if (bytes_out > DEF_TX_BUF_SIZE) {
        bytes_out = DEF_TX_BUF_SIZE;
        lcd_printf("Eth Frame truncated\r\n");
    }
    dst = (UINT8 *)tmpbuf;
    remains = bytes_out;
    while (pPBuf)
    {
        if (remains > pPBuf->len)
            memcpy(dst, pPBuf->payload, pPBuf->len);
        else {
            memcpy(dst, pPBuf->payload, remains);
            break;
        }
        remains -= pPBuf->len;
        dst += pPBuf->len;
        pPBuf = pPBuf->next;
    }
    Packet_Send((unsigned char *)tmpbuf, bytes_out);
    return ERR_OK;
}

void FM3_EMAC_LWIP_recv(struct netif *pNetIf)
{
    EMAC_DMA_RXDESC *prxdesc;
    INT32 recvd;
    struct pbuf *pPBuf;

    //hal_printf(".");
    GLOBAL_LOCK(irq);
    while ((prxdesc = RXDESC_Received()) != NULL) {
        recvd = prxdesc->RDES0_f.FL;
        pPBuf = pbuf_alloc(PBUF_RAW, recvd, PBUF_RAM);
        if (pPBuf == NULL) {
            hal_printf("Can't alloc pPBuf\r\n");
            break;
        }
        memcpy(pPBuf->payload, (void *)prxdesc->RDES2, recvd);
        prxdesc->RDES0_f.OWN = 1;
        pNetIf->input(pPBuf, pNetIf);
    }
#ifndef HW_INTERRUPT
    InterruptTaskCompletion.EnqueueDelta64(INT_PERIOD);
#endif
}
