////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2012 Kentaro Sekimoto  All rights reserved.
////////////////////////////////////////////////////////////////////////////

#include <tinyhal.h>
#include "..\FM3.h"

struct FM3_PWM_Driver
{
    static const UINT32 c_Channels = 8;

    static const GPIO_PIN c_Channel_0 = FM3__GPIO::P40; // TIOA00_0
    static const GPIO_PIN c_Channel_1 = FM3__GPIO::P41; // TIOA01_0
    static const GPIO_PIN c_Channel_2 = FM3__GPIO::P42; // TIOA02_0
    static const GPIO_PIN c_Channel_3 = FM3__GPIO::P43; // TIOA03_0
    static const GPIO_PIN c_Channel_4 = FM3__GPIO::P44; // TIOA04_0
    static const GPIO_PIN c_Channel_5 = FM3__GPIO::P45; // TIOA05_0
    static const GPIO_PIN c_Channel_6 = FM3__GPIO::PF3; // TIOA06_0
    static const GPIO_PIN c_Channel_7 = FM3__GPIO::P7C; // TIOA07_0

//    void FM3_PWM_Driver();
    BOOL    Initialize        ( PWM_CHANNEL channel );
    BOOL    Uninitialize      ( PWM_CHANNEL channel );
    BOOL    ApplyConfiguration( PWM_CHANNEL channel, GPIO_PIN pin, UINT32& period, UINT32& duration, BOOL invert );
    BOOL    Start             ( PWM_CHANNEL channel, GPIO_PIN pin );
    void    Stop              ( PWM_CHANNEL channel, GPIO_PIN pin );
    BOOL    Start             ( PWM_CHANNEL* channel, GPIO_PIN* pin, UINT32 count );
    void    Stop              ( PWM_CHANNEL* channel, GPIO_PIN* pin, UINT32 count );
    UINT32  Channels          ( );
    GPIO_PIN GetPinForChannel  ( PWM_CHANNEL channel );

    void EnablePin(PWM_CHANNEL channel);
    void DisablePin(PWM_CHANNEL channel);

    static const GPIO_PIN c_Pins[c_Channels];
    static const FM3_BT_PWM_TypeDef *c_PWM[c_Channels];
};

///////////////////////////////////////////////////////////////////////////////

//#if defined(ADS_LINKER_BUG__NOT_ALL_UNUSED_VARIABLES_ARE_REMOVED)
//#pragma arm section zidata = "g_FM3_PWM_Driver"
//#endif

const GPIO_PIN FM3_PWM_Driver::c_Pins[FM3_PWM_Driver::c_Channels] = {
    FM3__GPIO::P40,
    FM3__GPIO::P41,
    FM3__GPIO::P42,
    FM3__GPIO::P43,
    FM3__GPIO::P44,
    FM3__GPIO::P45,
    FM3__GPIO::PF3,
    FM3__GPIO::P7C
};

const FM3_BT_PWM_TypeDef *FM3_PWM_Driver::c_PWM[c_Channels] = {
    (FM3_BT_PWM_TypeDef *)(FM3_PERIPH_BASE + 0x25000UL),
    (FM3_BT_PWM_TypeDef *)(FM3_PERIPH_BASE + 0x25040UL),
    (FM3_BT_PWM_TypeDef *)(FM3_PERIPH_BASE + 0x25080UL),
    (FM3_BT_PWM_TypeDef *)(FM3_PERIPH_BASE + 0x250C0UL),
    (FM3_BT_PWM_TypeDef *)(FM3_PERIPH_BASE + 0x25200UL),
    (FM3_BT_PWM_TypeDef *)(FM3_PERIPH_BASE + 0x25240UL),
    (FM3_BT_PWM_TypeDef *)(FM3_PERIPH_BASE + 0x25280UL),
    (FM3_BT_PWM_TypeDef *)(FM3_PERIPH_BASE + 0x252C0UL)
};

FM3_PWM_Driver g_FM3_PWM_Driver;

//#if defined(ADS_LINKER_BUG__NOT_ALL_UNUSED_VARIABLES_ARE_REMOVED)
//#pragma arm section zidata
//#endif

BOOL PWM_Initialize(PWM_CHANNEL channel)
{
    return g_FM3_PWM_Driver.Initialize(channel);
}

BOOL PWM_Uninitialize(PWM_CHANNEL channel)
{
    return g_FM3_PWM_Driver.Uninitialize(channel);
}

BOOL PWM_ApplyConfiguration(PWM_CHANNEL channel, GPIO_PIN pin, UINT32& period, UINT32& duration, BOOL invert)
{
    return g_FM3_PWM_Driver.ApplyConfiguration(channel, pin, period, duration, invert);
}

BOOL PWM_Start(PWM_CHANNEL channel, GPIO_PIN pin)
{
    return g_FM3_PWM_Driver.Start(channel,pin);
}

void PWM_Stop(PWM_CHANNEL channel, GPIO_PIN pin)
{
    return g_FM3_PWM_Driver.Stop(channel,pin);
}

BOOL PWM_Start(PWM_CHANNEL* channel, GPIO_PIN* pin, UINT32 count)
{
    return g_FM3_PWM_Driver.Start(channel, pin, count);
}

void PWM_Stop(PWM_CHANNEL* channel, GPIO_PIN* pin, UINT32 count)
{
    return g_FM3_PWM_Driver.Stop(channel, pin, count);
}

UINT32 PWM_PWMChannels() 
{
    return g_FM3_PWM_Driver.Channels();
}

GPIO_PIN PWM_GetPinForChannel(PWM_CHANNEL channel)
{
    return g_FM3_PWM_Driver.GetPinForChannel(channel);
}

//--//

BOOL FM3_PWM_Driver::Initialize(PWM_CHANNEL channel)
{
    return TRUE;
}

BOOL FM3_PWM_Driver::Uninitialize(PWM_CHANNEL channel)
{
    return TRUE;
}

/*
 *  Register Bit Definition
 */
#define TMCR_STRG      0x0001U
#define TMCR_CTEN      0x0002U
#define TMCR_MDSE      0x0004U
#define TMCR_OSEL      0x0008U
#define TMCR_FMD_RST   0x0000U
#define TMCR_FMD_PWM   0x0010U
#define TMCR_EGS_DIS   0x0000U
#define TMCR_EGS_RISE  0x0100U
#define TMCR_EGS_FALL  0x0200U
#define TMCR_EGS_BOTH  0x0300U
#define TMCR_PMSK      0x0400U
#define TMCR_RTGEN     0x0800U
#define TMCR_CKS02_0   0x0000U
#define TMCR_CKS02_1   0x1000U
#define TMCR_CKS02_2   0x2000U
#define TMCR_CKS02_3   0x3000U
#define TMCR_CKS02_4   0x4000U
#define TMCR_CKS02_5   0x5000U
#define TMCR_CKS02_6   0x6000U
#define TMCR_CKS02_7   0x7000U

#define TMCR2_CKS3_0   0x00U
#define TMCR2_CKS3_1   0x01U

#define STC_UDIR       0x01U
#define STC_DTIR       0x02U
#define STC_TGIR       0x04U
#define STC_UDIE       0x10U
#define STC_DTIE       0x20U
#define STC_TGIE       0x40U
#define STC_IRQF       (STC_UDIR | STC_DTIR | STC_TGIR)

// TMCR
// b0: STRG
// b1: CTEN
// b2: MDSE
// b3: OSEL
// b4: FMD0
// b5: FMD1
// b6: FMD2
// b7: reserved
// b8: EGS0
// b9: EGS1
// b10:PMSK
// b11:RTGEN
// b12:CKS0
// b13:CKS1
// b14:CKS2
// b15 reserved
//#define DEBUG_PWM
#define PWM_CLKDEV        16        // 72MZ / CLKDEV / 256
BOOL FM3_PWM_Driver::ApplyConfiguration(PWM_CHANNEL channel, GPIO_PIN pin, UINT32 & period,UINT32 & duration,BOOL invert)
{
    if ((UINT32)channel >= c_Channels)
        return FALSE;
    UINT32 flag_started = FALSE;
    UINT32 period_ticks   = CPU_MicrosecondsToSystemClocks(period) / (PWM_CLKDEV * CLKDEV);
    UINT32 duration_ticks = CPU_MicrosecondsToSystemClocks(duration) / (PWM_CLKDEV * CLKDEV);
#ifdef DEBUG_PWM
    lcd_printf("C/D/P %d/%04x/%04x\r\n", (UINT32)channel, (UINT16)duration_ticks, (UINT16)period_ticks);
#endif
    FM3_BT_PWM_TypeDef *pwm = (FM3_BT_PWM_TypeDef *)c_PWM[(UINT32)channel];
    if ((UINT32)channel < 6) {
        FM3_GPIO->PFR4 |= (1 << (UINT32)channel);    // peripheral mode
        FM3_GPIO->PZR4 |= (1 << (UINT32)channel);    // open drain mode
    } else if ((UINT32)channel == 6) {
        FM3_GPIO->PFRF |= 0x00000008;    // PF3 peripheral mode
        FM3_GPIO->PZRF |= 0x00000008;    // PF3 open drain mode
    } else {
        FM3_GPIO->PFR7 |= 0x00001000;    // P7C periphearl mode
        FM3_GPIO->PZR7 |= 0x00001000;    // P7C open drain mode
    }
    if ((UINT32)channel < 4)
        FM3_GPIO->EPFR04 |= (0x00000004 << ((UINT32)channel * 8));
    else
        FM3_GPIO->EPFR05 |= (0x00000004 << (((UINT32)channel-4) * 8));
    if ((pwm->TMCR & TMCR_CTEN) != 0) {
        flag_started = TRUE;
        pwm->TMCR &= ~(TMCR_CTEN);
    }
    if (invert)
        pwm->TMCR = 0x2814;    // CLK/16, PWM mode, continuous
    else
        pwm->TMCR = 0x2810;    // CLK/16, PWM mode, continuous
    pwm->TMCR2 = 0x00;
    pwm->STC = 0;            // CLK/16
    pwm->PCSR = (UINT16)period_ticks;
    pwm->PDUT = (UINT16)duration_ticks;
    if (flag_started == TRUE) {
        pwm->TMCR |= (TMCR_CTEN | TMCR_STRG);
    }
    return TRUE;
}

void FM3_PWM_Driver::EnablePin(PWM_CHANNEL channel)
{
//    CPU_GPIO_DisablePin(g_FM3_PWM_Driver.c_Pins[channel], RESISTOR_DISABLED, 0, GPIO_ALT_MODE_2);
    FM3_BT_PWM_TypeDef *pwm = (FM3_BT_PWM_TypeDef *)c_PWM[(UINT32)channel];
    pwm->TMCR |= (TMCR_CTEN | TMCR_STRG);
}

void FM3_PWM_Driver::DisablePin(PWM_CHANNEL channel)
{
    FM3_BT_PWM_TypeDef *pwm = (FM3_BT_PWM_TypeDef *)c_PWM[(UINT32)channel];
    pwm->TMCR &= ~(TMCR_CTEN | TMCR_STRG);
//    CPU_GPIO_DisablePin(g_FM3_PWM_Driver.c_Pins[channel], RESISTOR_PULLDOWN, 0, GPIO_ALT_PRIMARY);
}

BOOL FM3_PWM_Driver::Start(PWM_CHANNEL channel, GPIO_PIN pin)
{
    EnablePin(channel);
    return TRUE;
}

void FM3_PWM_Driver::Stop(PWM_CHANNEL channel, GPIO_PIN pin)
{    
    DisablePin(channel);
}

BOOL FM3_PWM_Driver::Start(PWM_CHANNEL* channel, GPIO_PIN* pin, UINT32 count)
{
    return TRUE;
}

void FM3_PWM_Driver::Stop(PWM_CHANNEL* channel, GPIO_PIN* pin, UINT32 count)
{
}

UINT32 FM3_PWM_Driver::Channels()
{
    return c_Channels;
}

GPIO_PIN FM3_PWM_Driver::GetPinForChannel(PWM_CHANNEL channel)
{
    if ((UINT32)channel >= c_Channels)
        return GPIO_PIN_NONE;
    return (GPIO_PIN)g_FM3_PWM_Driver.c_Pins[channel];
}
