////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Microsoft Corporation.  All rights reserved.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Portions Copyright (c) 2012 Kentaro Sekimoto  All rights reserved.
////////////////////////////////////////////////////////////////////////////

#include <tinyhal.h>
#include "..\FM3.h"
#include "FM3_touchpanel_driver.h"

//--//
struct FM3_TouchPanel_Driver g_FM3_TouchPanel_Driver;

BOOL FM3_TouchPanel_Driver::CalibrationPointGet(TOUCH_PANEL_CALIBRATION_POINT *pTCP)
{
	INT32   cDisplayWidth  = pTCP->cDisplayWidth;
	INT32   cDisplayHeight = pTCP->cDisplayHeight;
	int CalibrationRadiusX = cDisplayWidth  / 20;
	int CalibrationRadiusY = cDisplayHeight / 20;
	switch (pTCP -> PointNumber)
	{
	case    0:
		pTCP->CalibrationX = cDisplayWidth  / 2;
		pTCP->CalibrationY = cDisplayHeight / 2;
		break;

	case    1:
		pTCP->CalibrationX = CalibrationRadiusX * 2;
		pTCP->CalibrationY = CalibrationRadiusY * 2;
		break;

	case    2:
		pTCP->CalibrationX = CalibrationRadiusX * 2;
		pTCP->CalibrationY = cDisplayHeight - CalibrationRadiusY * 2;
		break;

	case    3:
		pTCP->CalibrationX = cDisplayWidth  - CalibrationRadiusX * 2;
		pTCP->CalibrationY = cDisplayHeight - CalibrationRadiusY * 2;
		break;

	case    4:
		pTCP->CalibrationX = cDisplayWidth - CalibrationRadiusX * 2;
		pTCP->CalibrationY = CalibrationRadiusY * 2;
		break;

	default:
		pTCP->CalibrationX = cDisplayWidth  / 2;
		pTCP->CalibrationY = cDisplayHeight / 2;

		return FALSE;
	}
	return TRUE;
}

HRESULT FM3_TouchPanel_Driver::GetDeviceCaps(unsigned int iIndex, void* lpOutput)
{
	if ( lpOutput == NULL )
	{
		return FALSE;
	}
	switch ( iIndex )
	{
	case TOUCH_PANEL_SAMPLE_RATE_ID:
	{
		TOUCH_PANEL_SAMPLE_RATE   *pTSR = (TOUCH_PANEL_SAMPLE_RATE*)lpOutput;

		pTSR->SamplesPerSecondLow       = g_TouchPanel_Sampling_Settings.SampleRate.SamplesPerSecondLow;
		pTSR->SamplesPerSecondHigh      = g_TouchPanel_Sampling_Settings.SampleRate.SamplesPerSecondHigh;
		pTSR->CurrentSampleRateSetting  = g_TouchPanel_Sampling_Settings.SampleRate.CurrentSampleRateSetting;
		pTSR->MaxTimeForMoveEvent_ticks = g_TouchPanel_Sampling_Settings.SampleRate.MaxTimeForMoveEvent_ticks;
	}
	break;

	case TOUCH_PANEL_CALIBRATION_POINT_COUNT_ID:
	{
		TOUCH_PANEL_CALIBRATION_POINT_COUNT *pTCPC = (TOUCH_PANEL_CALIBRATION_POINT_COUNT*)lpOutput;

		pTCPC->flags              = 0;
		pTCPC->cCalibrationPoints = 5;
	}
	break;

	case TOUCH_PANEL_CALIBRATION_POINT_ID:
		return(CalibrationPointGet((TOUCH_PANEL_CALIBRATION_POINT*)lpOutput));

	default:
		return FALSE;
	}
	return TRUE;
}

void FM3_TouchPanel_Driver::GetPoint(TOUCH_PANEL_SAMPLE_FLAGS* pTipState, int* pSource, int* pUnCalX, int* pUnCalY)
{
	static BOOL stylusDown = FALSE;
	INT32 validReadCount = 0;
    *pTipState = 0;
    *pUnCalX = 0;
    *pUnCalY = 0;
    *pSource = 0;

	if (stylusDown)
	{
		*pTipState |= TouchSamplePreviousDownFlag;
	}

	if (g_FM3_TouchPanel_Driver.m_global_touch)
	{
		*pTipState |= TouchSampleValidFlag;
		*pUnCalX = g_FM3_TouchPanel_Driver.m_global_inX;
		*pUnCalY = g_FM3_TouchPanel_Driver.m_global_inY;
		*pTipState |= TouchSampleDownFlag;
		stylusDown = true;
	}
	else
	{
		*pUnCalX = -1;
		*pUnCalY = -1;
		stylusDown = false;
	}
}

void FM3_TouchPanel_Driver::TouchCompletion(void* arg)
{
	g_FM3_TouchPanel_Driver.Poll();
    if (!g_FM3_TouchPanel_Driver.m_touchCompletion.IsLinked()) {
    	g_FM3_TouchPanel_Driver.m_touchCompletion.EnqueueDelta(g_FM3_TouchPanel_Driver.m_samplingTimespan);
    } else
    {
    	g_FM3_TouchPanel_Driver.m_touchCompletion.Abort();
    }
}

// CN2-01 PB0 XIN
// CN2-02 PB1 YIN
// CN2-33 P36 XEN
// CN2-34 P37 YEN
// CN2-35 P91 YHL
// CN2-36 P92 LCD_POWER
BOOL FM3_TouchPanel_Driver::Initialize(GPIO_INTERRUPT_SERVICE_ROUTINE touchIsrProc)
{
	FM3_GPIO->PFR3 &= ~0x000000C0;	// P36, P37: GPIO
	FM3_GPIO->DDR3 |= 0x000000C0;	// P36, P37: output
	FM3_GPIO->PFR9 &= ~0x00000006;	// P91, P92: GPIO
	FM3_GPIO->DDR9 |= 0x00000006;	// P91, P92: output
	FM3_GPIO->PFRF &= ~0x00000001;	// PF1: GPIO
	FM3_GPIO->DDRF &= ~0x00000001;	// PF1: input
	FM3_GPIO->PCRF |= 0x00000001;	// PF1: pull up
	// A-D configuration
	// AN16 - ADC channel 1 for XIN
	// AN17 - ADC channel 2 for YIN
	// ADC channel 0 for other NETMF
	FM3_GPIO->ADE |= 0x00030000;	// AN16, AN17: Analog
	FM3_GPIO->PFRB  |= 0x00000003;	// AN16(PB0), AN17(PB1): Analog
	FM3_ADC1->ADCT = 3;
	FM3_ADC1->ADCEN = ADCEN_ENBL;
    while((FM3_ADC1->ADCEN & ADCEN_READY) != ADCEN_READY) ;
    // AN16 - channel 1
    FM3_ADC1->SCIS2 = (1 << 0);
	FM3_ADC1->ADST0 = 0xA9;	// PCLK 72MHz=14ns -> 14ns x 64 x 10 = 8960ns
	FM3_ADC1->ADSS2 = (0 << 0);
	FM3_ADC1->CMPCR = 16;
	FM3_ADC1->ADSR = ADSR_FDAS;		// ADSTP(R/W):0, FDAS(R/W):1(LSB), , , , PCNS(R):0, PCS(R):0, SCS(R):0
	FM3_ADC1->ADCR = 0;				// SCIF:0, PCIF:0, CMPIF:0, , SCIE:0 , PCIE:0 , CMPIE:0, OVRIE:0
	FM3_ADC1->SCCR |= SCCR_SFCLR;
	FM3_ADC1->SCCR &= ~SCCR_RPT;
	// AN17 - channel 2
	FM3_ADC2->ADCEN = ADCEN_ENBL;
    while((FM3_ADC2->ADCEN & ADCEN_READY) != ADCEN_READY) ;
	FM3_ADC2->SCIS2 = (1 << 1);
	FM3_ADC2->ADST0 = 0xA9;	// PCLK 72MHz=14ns -> 14ns x 64 x 10 = 8960ns
	FM3_ADC2->ADSS2 = (0 << 1);
	FM3_ADC2->CMPCR = 17;
	FM3_ADC2->ADSR = ADSR_FDAS;		// ADSTP(R/W):0, FDAS(R/W):1(LSB), , , , PCNS(R):0, PCS(R):0, SCS(R):0
	FM3_ADC2->ADCR = 0;				// SCIF:0, PCIF:0, CMPIF:0, , SCIE:0 , PCIE:0 , CMPIE:0, OVRIE:0
	FM3_ADC2->SCCR |= SCCR_SFCLR;
	FM3_ADC2->SCCR &= ~SCCR_RPT;
	g_FM3_TouchPanel_Driver.m_x_min = ((X_UL + X_DL) / 2);
	g_FM3_TouchPanel_Driver.m_x_max = ((X_UR + X_DR) / 2);
	g_FM3_TouchPanel_Driver.m_y_min = ((Y_UL + Y_UR) / 2);
	g_FM3_TouchPanel_Driver.m_y_max = ((Y_DL + Y_DR) / 2);
	g_FM3_TouchPanel_Driver.m_global_inX = 0;
	g_FM3_TouchPanel_Driver.m_global_inY = 0;
	g_FM3_TouchPanel_Driver.m_global_touch = 0;
	g_FM3_TouchPanel_Driver.m_samplingTimespan = FM3_TOUCHPANEL_POLL_TIMESPAN;
    g_FM3_TouchPanel_Driver.m_touchCompletion.Initialize();
    g_FM3_TouchPanel_Driver.m_touchCompletion.InitializeForISR(FM3_TouchPanel_Driver::TouchCompletion, NULL);
	g_FM3_TouchPanel_Driver.m_touchCompletion.EnqueueDelta(g_FM3_TouchPanel_Driver.m_samplingTimespan);
    if (!::CPU_GPIO_EnableInputPin2(
            0xF1,
            FALSE,
            touchIsrProc,
            NULL,
            GPIO_INT_EDGE_BOTH,
            RESISTOR_PULLUP))
    {
        return FALSE;
    }
    return TRUE;
}

BOOL FM3_TouchPanel_Driver::Uninitialize()
{
    if (g_FM3_TouchPanel_Driver.m_touchCompletion.IsLinked())
    {
        g_FM3_TouchPanel_Driver.m_touchCompletion.Abort();
    }
    return TRUE;
}

void FM3_TouchPanel_Driver::Poll()
{
	static UINT32 t_state;
	UINT32 value;

	switch(t_state) {
	case 0:
		SENSE_ON;
		t_state++;
		break;
	case 1:
		TOUCH_CHK_ON;
		if (TOUCH_CHK)
			t_state++;
		else {
			IDLE;
			g_FM3_TouchPanel_Driver.m_global_touch = 0;
			t_state = 0;
		}
		TOUCH_CHK_OFF;
		break;
	case 2:	// skip for chattering
	case 3:
		t_state++;
		break;
	case 4:
		TOUCH_CHK_ON;
		if (TOUCH_CHK) {
			TOUCH_CHK_OFF;
			SENSE_Y_ON;
			t_state++;
		} else {
			TOUCH_CHK_OFF;
			IDLE;
			g_FM3_TouchPanel_Driver.m_global_touch = 0;
			t_state = 0;
		}
		break;
	case 5:	// start A-D for getting Y axis
		FM3_ADC2->SCCR |= SCCR_SSTR;	// A/D start
		t_state++;
		break;
	case 6:	// sense power on
		SENSE_ON;
		t_state++;
		break;
	case 7:
		TOUCH_CHK_ON;
		if (TOUCH_CHK) {
			TOUCH_CHK_OFF;
			SENSE_X_ON;
			t_state++;
		} else {
			TOUCH_CHK_OFF;
			IDLE;
			g_FM3_TouchPanel_Driver.m_global_touch = 0;
			t_state = 0;
		}
		break;
	case 8:	// start A-D for sensing X axis
		FM3_ADC1->SCCR |= SCCR_SSTR;	// A/D start
		t_state++;
		break;
	case 9:	// sense power on
		SENSE_ON;
		t_state++;
		break;
	case 10:	//  skip for chattering
	case 11:
		t_state++;
		break;
	case 12:
		TOUCH_CHK_ON;
		if (TOUCH_CHK) {
			TOUCH_CHK_OFF;
			g_FM3_TouchPanel_Driver.m_global_touch = 1;
			do {
				value = FM3_ADC1->SCFD;
			} while (((value & SCFDL_INVL) != 0) || (16 != (value & 0x1F)));
			g_FM3_TouchPanel_Driver.m_global_inY = (UINT16)(value >> 16);
			do {
				value = FM3_ADC2->SCFD;
			} while (((value & SCFDL_INVL) != 0) || (17 != (value & 0x1F)));
			g_FM3_TouchPanel_Driver.m_global_inX = (UINT16)(value >> 16);
		} else
			TOUCH_CHK_OFF;
		t_state = 0;
		FM3_ADC1->SCCR = (FM3_ADC1->SCCR | SCCR_SFCLR) & ~(SCCR_RPT | SCCR_SOVR);	// A/D stop
		FM3_ADC2->SCCR = (FM3_ADC2->SCCR | SCCR_SFCLR) & ~(SCCR_RPT | SCCR_SOVR);	// A/D stop
		break;
	default:
		t_state = 0;
		break;
	}
}
