/*
 * xtsnprintf.c
 *
 *  Created on: 2011/03/21
 *      Author: Kentaro Sekimoto
 */
//#include <stdio.h>
#include <stdarg.h>

#define MORETHAN10	1
#define	MAX_DIGIT	20
const char sdigit_str[] = "0123456789abcdef";
const char bdigit_str[] = "0123456789ABCDEF";

// xput_s
// put a string into buffer and fill with spaces
unsigned int xput_s(char *buf, char *s, int n, unsigned int buf_size)
{
	char *p;
	unsigned int size = 0;
	if (n > 0) {
		n -= strlen(s);	// decreased by string length
		if (n > 0) {
			memset(buf, ' ', n);
			size += n;
			buf += n;
			buf_size -= n;
		}
	}
	for (p = s; *p != 0; p++) {
		if (buf_size-- == 0)
			return -1;
		*buf++ = *p;
		size++;
	}
	return size;
}

// xput_c
// put a character into buffer and fill with spaces
unsigned int xput_c(char *buf, char c, int n, unsigned int buf_size)
{
	unsigned int size = 0;
	if (buf_size > 0) {
		if (n > 0) {
			n--;	// one character
			if (n > buf_size)
				n = buf_size;
			memset(buf, ' ', n);
			buf += n;
			size += n;
		}
		*buf++ = c;
		size++;
	}
	return size;
}

// xput_n
// put a decimal number into buffer
unsigned int xput_n(char *buf, char *fmt, unsigned long long vv, int digits, unsigned int zeroflag, unsigned int buf_size)
{
	char tmp[MAX_DIGIT];

	char *p, *t, z;
	unsigned int base;
	unsigned int cap = 0;
	unsigned int size = 0;

	if (zeroflag)
		z = '0';
	else
		z = ' ';
	switch (*fmt) {
	case 'b':
		base = 2;
		break;
	case 'd':
		base = 10;
		break;
	case 'u':
		base = 10;
		break;
	case 'o':
		base = 8;
		break;
	case 'x':
		base = 16;
		break;
	case 'D':
		base = 10;
		cap = 1;
		break;
	case 'O':
		base = 8;
		cap = 1;
		break;
	case 'X':
		base = 16;
		cap = 1;
		break;
	default:
		return size;
	}
	p = tmp;
	t = (char *)sdigit_str;
	if (cap)
		t = (char *)bdigit_str;
	do {
		digits--;
		*p++ = t[vv % base];
		vv /= base;
	} while (vv != 0);
	for (; digits > 0; digits--) {
		if (buf_size == 0)
			return -1;
		buf_size--;
		*buf++ = z;
		size++;
	}
	while (p > tmp) {
		if (buf_size == 0)
			return -1;
		buf_size--;
		*buf++ = *--p;
		size++;
	}
	return size;
}

// xput_f
// put a float number into buffer
int xput_f(char *buf, char *fmt, double f, int n, int m, int zeroflag, unsigned int buf_size)
{
	char tmp[MAX_DIGIT];

	long l;
	long base;
	char pc, z;
	int i;
	int size = 0;
	char *p, *t;

	if (zeroflag)
		z = '0';
	else
		z = ' ';
	base = 10L;
	if (f < 0) {
		pc = '-';
		f = -f;
	}
	for (i = 0; i < m; i++)
		f *= 10.0;
	l = (long)f;
	p = tmp;
	t = (char *)sdigit_str;
	do {
		if (m == 0)
			*p++ = '.';
		n--;
		m--;
		*p++ = t[l % base];
		l /= base;
	} while (l != 0);
	for (; n > 0; n--) {
		if (buf_size == 0)
			return -1;
		buf_size--;
		*buf++ = z;
		size++;
	}
	while (p > tmp) {
		if (buf_size == 0)
			return -1;
		buf_size--;
		*buf++ = *--p;
		size++;
	}
	return size;
}

// vxsnprintf
int vsnprintf(char *buf, int buf_size, char *fmt, va_list args)
{
	char p;
	int n, m;
	unsigned long v;
	unsigned long long vv;
	int size;
	int len = 0;
	int zeroflag = 0;
	double f;

	for (; *fmt != '\0'; fmt++) {
		p = *fmt;
		if (p == '%') {
			size = 0;
			fmt++;
			p = *fmt;
			if (p == '0')
				zeroflag = 1;
			if (isdigit(p)) {
				n = *fmt++ - '0';
#if MORETHAN10
				while (isdigit(*fmt)) n = n * 10 + (*fmt++ - '0');
#endif
				if (*fmt == '.') {
					fmt++;
					if (isdigit(*fmt)) {
						m = *fmt++ - '0';
#if MORETHAN10
						while (isdigit(*fmt)) m = m * 10 + (*fmt++ - '0');
					}
				}
				p = *fmt;
#endif
			} else
				n = -1;

			switch (p) {
			case 'l':
				{
					char p1 = *(fmt+1);
					char p2 = *(fmt+2);
					if ((p1 == 'd') || (p1 == 'u') || (p1 == 'x') ||  (p1 == 'X')) {
						vv = (unsigned long long)va_arg(args, long);
						fmt += 1;
						size = xput_n(buf, fmt, vv, n, zeroflag, buf_size);

					} else if ((p1 == 'l') && ((p2 == 'd') || (p2 == 'u') || (p2 == 'x') ||  (p2 == 'X'))) {
						vv = (unsigned long long)va_arg(args, long long);
						fmt += 2;
						size = xput_n(buf, fmt, vv, n, zeroflag, buf_size);
					}
				}
				break;
			case 's':
				v = (unsigned long)va_arg(args, long);
				size = xput_s(buf, (char *)v, n, buf_size);
				break;
			case 'c':
				v = (unsigned long)va_arg(args, long);
				size = xput_c(buf, (char)v, n, buf_size);
				break;
			case 'b':
			case 'd':
			case 'u':
			case 'o':
			case 'x':
			case 'e':
			case 'O':
			case 'X':
				v = (unsigned long)va_arg(args, long);
				size = xput_n(buf, fmt, (long long)v, n, zeroflag, buf_size);
				break;
			case 'f':
				f = (float)va_arg(args, double);
				size = xput_f(buf, fmt, f, n, m, zeroflag, buf_size);
				break;
			}
			if (size < 0)
				return -1;
			if (size) {
				buf += size;
				len += size;
				buf_size -= size;
			}
		}
		else {
			if (buf_size == 0)
				return -1;
			*buf++ = p;
			len++;
			buf_size--;
		}
	}
	*buf++ = '\0';
	return len;
}
