////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2010-2011 Kentaro Sekimoto  All rights reserved.
////////////////////////////////////////////////////////////////////////////
#include <tinyhal.h>
#include "..\..\..\..\DeviceCode\Targets\Native\FM3\DeviceCode\FM3.h"
#include "..\..\..\..\DeviceCode\Drivers\FS\FAT\FAT_FS.h"

//HAL_DECLARE_NULL_HEAP();
HAL_DECLARE_CUSTOM_HEAP(SimpleHeap_Allocate, SimpleHeap_Release,SimpleHeap_ReAllocate);

void PostManagedEvent( UINT8 category, UINT8 subCategory, UINT16 data1, UINT32 data2 )
{
}

extern struct BlockStorageDevice  g_SD_BS;
extern struct IBlockStorageDevice g_SD_BS_DeviceTable;
extern struct SD_BLOCK_CONFIG g_SD_BS_Config;

const char nameSpace[] = "SD";
const WCHAR path[] = {'\\', 'S', 'D', 'T', 'E', 'S', 'T', '.', 'T', 'X', 'T', '\0'};
FileSystemVolume *pFSVolume = (FileSystemVolume *)NULL;
#define SD_BUF_SIZE	128
char rbuf[SD_BUF_SIZE+1];
char wbuf[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ-abcdefghijklmnopqrstuvwxyz";

BOOL SDMount(const char *nameSpace)
{
	if (BlockStorageList::AddDevice(&g_SD_BS, &g_SD_BS_DeviceTable, &g_SD_BS_Config, TRUE)) {
		FS_MountVolume(nameSpace, 0, 0, &g_SD_BS);
		pFSVolume = FileSystemVolumeList::FindVolume(nameSpace, hal_strlen_s((char *)nameSpace));
		debug_printf("SDMount: OK\r\n");
		return TRUE;
	} else {
		debug_printf("SDMount: NG\r\n");
		return FALSE;
	}
}

BOOL SDOpenFile(UINT32 *handle, LPCWSTR path)
{
	BOOL ret = FALSE;
	if (pFSVolume) {
		FAT_FS_Driver::Open(&pFSVolume->m_volumeId, (LPCWSTR)path, handle);
		if (*handle)
			ret = TRUE;
	} else
		debug_printf("SDOpenFile: volume not mounted\r\n");
	if (ret)
		debug_printf("SDOpenFile: OK\r\n");
	else
		debug_printf("SDOpenFile: NG\r\n");
	return ret;
}

BOOL SDReadFile(LPCWSTR path)
{
	UINT32 handle;
	BOOL ret = FALSE;
	int readCount = 0;
	if (SDOpenFile(&handle, path)) {
		HRESULT hResult = FAT_FS_Driver::Read(handle, (BYTE *)rbuf, SD_BUF_SIZE, &readCount);
		if (hResult == S_OK) {
			if (readCount < 1)
				debug_printf("SDReadFile: file is empty\r\n");
			else {
				rbuf[readCount] = '\0';
				debug_printf("SDReadFile: read data: %s\r\n", rbuf);
			}
			ret = TRUE;
		}
		FAT_FS_Driver::Close(handle);
	}
	if (ret)
		debug_printf("SDReadFile: OK\r\n");
	else
		debug_printf("SDReadFile: NG\r\n");
	return ret;
}

BOOL SDWriteFile(LPCWSTR path)
{
	UINT32 handle;
	BOOL ret = FALSE;
	int writeLen = hal_strlen_s(wbuf);
	int writeCount = 0;
	if (SDOpenFile(&handle, path)) {
		writeCount = 0;
		HRESULT hResult = FAT_FS_Driver::Write(handle, (BYTE*)wbuf, writeLen, &writeCount);
		if ((hResult == S_OK) && (writeCount == writeLen))
			ret = TRUE;
		FAT_FS_Driver::Close(handle);
	}
	if (ret)
		debug_printf("SDWriteFile: OK\r\n");
	else
		debug_printf("SDWriteFile: NG\r\n");
	return ret;
}

BOOL SDDeleteFile(LPCWSTR path)
{
	UINT32 handle;
	HRESULT hResult;
	BOOL ret = FALSE;
	if (SDOpenFile(&handle, path)) {
		if(!handle)
			FAT_FS_Driver::Close(handle);
		hResult = FAT_FS_Driver::Delete(&pFSVolume->m_volumeId, (LPCWSTR)path);
		if (hResult == S_OK) {
			debug_printf("SDDeleteFile: Delete OK\r\n");
			ret = TRUE;
		} else if (hResult == CLR_E_FILE_IO)
			debug_printf("SDDeleteFile: IO Err\r\n");
		else if(hResult == CLR_E_FILE_NOT_FOUND)
			debug_printf("SDDefileFile: File not fount\r\n");
		else if(hResult == CLR_E_UNAUTHORIZED_ACCESS)
			debug_printf("SDDeleteFile: Unauthorized\r\n");
		hResult = FAT_FS_Driver::FlushAll(&pFSVolume->m_volumeId);
		if (hResult == S_OK)
			debug_printf("SDDeleteFile: Flash OK\r\n");
		else {
			debug_printf("SDDeleteFile: Flash NG\r\n");
			ret = FALSE;
		}
	} else
		debug_printf("SDDeleteFile: Delete NG\r\n");
	if (ret)
		debug_printf("SDDeleteFile: OK\r\n");
	else
		debug_printf("SDDeleteFile: NG\r\n");
	return ret;
}

void TestSD(void)
{
	UINT32 handle;
	debug_printf("TestSD\r\n");
	if (SDMount(nameSpace)) {
		SDWriteFile(path);
		SDReadFile(path);
		SDDeleteFile(path);
	}
}

void ApplicationEntryPoint()
{
	HalSystemConfig.stdio = COM1;
	HalSystemConfig.DebugTextPort = COM1;
	DebuggerPort_Initialize(HalSystemConfig.DebugTextPort);
	TestSD();
	while (true);
}
