////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Microsoft Corporation.  All rights reserved.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Portions Copyright (c) 2012 Kentaro Sekimoto  All rights reserved.
////////////////////////////////////////////////////////////////////////////

#include <tinyhal.h>
#include "..\..\..\..\DeviceCode\Targets\Native\FM3\DeviceCode\FM3.h"
#include "CQ_FRK_FM3_TouchPanel_Driver.h"

//--//
//#define DEBUG_TOUCHPANEL
struct CQ_FRK_FM3_TouchPanel_Driver g_CQ_FRK_FM3_TouchPanel_Driver;

BOOL CQ_FRK_FM3_TouchPanel_Driver::CalibrationPointGet(TOUCH_PANEL_CALIBRATION_POINT *pTCP)
{
    INT32   cDisplayWidth  = pTCP->cDisplayWidth;
    INT32   cDisplayHeight = pTCP->cDisplayHeight;
    int CalibrationRadiusX = cDisplayWidth  / 20;
    int CalibrationRadiusY = cDisplayHeight / 20;
    switch (pTCP -> PointNumber)
    {
    case    0:
        pTCP->CalibrationX = cDisplayWidth  / 2;
        pTCP->CalibrationY = cDisplayHeight / 2;
        break;

    case    1:
        pTCP->CalibrationX = CalibrationRadiusX * 2;
        pTCP->CalibrationY = CalibrationRadiusY * 2;
        break;

    case    2:
        pTCP->CalibrationX = CalibrationRadiusX * 2;
        pTCP->CalibrationY = cDisplayHeight - CalibrationRadiusY * 2;
        break;

    case    3:
        pTCP->CalibrationX = cDisplayWidth  - CalibrationRadiusX * 2;
        pTCP->CalibrationY = cDisplayHeight - CalibrationRadiusY * 2;
        break;

    case    4:
        pTCP->CalibrationX = cDisplayWidth - CalibrationRadiusX * 2;
        pTCP->CalibrationY = CalibrationRadiusY * 2;
        break;

    default:
        pTCP->CalibrationX = cDisplayWidth  / 2;
        pTCP->CalibrationY = cDisplayHeight / 2;

        return FALSE;
    }
    return TRUE;
}

HRESULT CQ_FRK_FM3_TouchPanel_Driver::GetDeviceCaps(unsigned int iIndex, void* lpOutput)
{
    if ( lpOutput == NULL )
    {
        return FALSE;
    }
    switch ( iIndex )
    {
    case TOUCH_PANEL_SAMPLE_RATE_ID:
    {
        TOUCH_PANEL_SAMPLE_RATE   *pTSR = (TOUCH_PANEL_SAMPLE_RATE*)lpOutput;

        pTSR->SamplesPerSecondLow       = g_TouchPanel_Sampling_Settings.SampleRate.SamplesPerSecondLow;
        pTSR->SamplesPerSecondHigh      = g_TouchPanel_Sampling_Settings.SampleRate.SamplesPerSecondHigh;
        pTSR->CurrentSampleRateSetting  = g_TouchPanel_Sampling_Settings.SampleRate.CurrentSampleRateSetting;
        pTSR->MaxTimeForMoveEvent_ticks = g_TouchPanel_Sampling_Settings.SampleRate.MaxTimeForMoveEvent_ticks;
    }
    break;

    case TOUCH_PANEL_CALIBRATION_POINT_COUNT_ID:
    {
        TOUCH_PANEL_CALIBRATION_POINT_COUNT *pTCPC = (TOUCH_PANEL_CALIBRATION_POINT_COUNT*)lpOutput;

        pTCPC->flags              = 0;
        pTCPC->cCalibrationPoints = 5;
    }
    break;

    case TOUCH_PANEL_CALIBRATION_POINT_ID:
        return(CalibrationPointGet((TOUCH_PANEL_CALIBRATION_POINT*)lpOutput));

    default:
        return FALSE;
    }
    return TRUE;
}

void CQ_FRK_FM3_TouchPanel_Driver::GetPoint(TOUCH_PANEL_SAMPLE_FLAGS* pTipState, int* pSource, int* pUnCalX, int* pUnCalY)
{
    static BOOL stylusDown = FALSE;
    INT32 validReadCount = 0;
    *pTipState = 0;
    *pUnCalX = 0;
    *pUnCalY = 0;
    *pSource = 0;

    if (g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inX >= 3700)
        validReadCount = 0;
    else
        validReadCount = 1;
    if (stylusDown)
        *pTipState |= TouchSamplePreviousDownFlag;
    if ((validReadCount == 1) &&
        (g_CQ_FRK_FM3_TouchPanel_Driver.m_global_touch == 1)) {
        *pTipState |= TouchSampleValidFlag;
        *pUnCalX = g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inY;  // intentionally changed x <-> y
        *pUnCalY = g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inX;  // intentionally changed x <-> y
        *pTipState |= TouchSampleDownFlag;
        stylusDown = true;
    } else {
        *pUnCalX = -1;
        *pUnCalY = -1;
        stylusDown = false;
    }
}

void CQ_FRK_FM3_TouchPanel_Driver::TouchCompletion(void* arg)
{
    g_CQ_FRK_FM3_TouchPanel_Driver.Poll();
    if (!g_CQ_FRK_FM3_TouchPanel_Driver.m_touchCompletion.IsLinked()) {
        g_CQ_FRK_FM3_TouchPanel_Driver.m_touchCompletion.EnqueueDelta(g_CQ_FRK_FM3_TouchPanel_Driver.m_samplingTimespan);
    } else
    {
        g_CQ_FRK_FM3_TouchPanel_Driver.m_touchCompletion.Abort();
    }
}

// CN2-01 PB0 XIN
// CN2-02 PB1 YIN
// CN2-33 P36 XEN
// CN2-34 P37 YEN
// CN2-35 P91 YHL
// CN2-36 P92 LCD_POWER
BOOL CQ_FRK_FM3_TouchPanel_Driver::Initialize(GPIO_INTERRUPT_SERVICE_ROUTINE touchIsrProc)
{
    // A-D configuration
    // AN16 - ADC channel 1 for XIN
    // AN17 - ADC channel 2 for YIN
    // ADC channel 0 for other NETMF
    FM3_GPIO->ADE &= 0x007F0000;
    FM3_GPIO->PFRB &= ~0x0000007F;
    FM3_ADC1->ADCEN = ADCEN_ENBL;
    while((FM3_ADC1->ADCEN & ADCEN_READY) != ADCEN_READY) ;
    // AN16 - channel 1
    FM3_ADC1->SCIS2 = (1 << (16-16));
    FM3_ADC1->ADST0 = 0xA9;    // PCLK 72MHz=14ns -> 14ns x 64 x 10 = 8960ns
    FM3_ADC1->ADSS2 = (0 << 0);
    FM3_ADC1->CMPCR = 16;
    FM3_ADC1->ADSR = ADSR_FDAS;        // ADSTP(R/W):0, FDAS(R/W):1(LSB), , , , PCNS(R):0, PCS(R):0, SCS(R):0
    FM3_ADC1->ADCR = 0;                // SCIF:0, PCIF:0, CMPIF:0, , SCIE:0 , PCIE:0 , CMPIE:0, OVRIE:0
    FM3_ADC1->SCCR |= SCCR_SFCLR;
    FM3_ADC1->SCCR &= ~SCCR_RPT;
    // AN17 - channel 2
    FM3_ADC2->ADCEN = ADCEN_ENBL;
    while((FM3_ADC2->ADCEN & ADCEN_READY) != ADCEN_READY) ;
    FM3_ADC2->SCIS2 = (1 << (17-16));
    FM3_ADC2->ADST0 = 0xA9;    // PCLK 72MHz=14ns -> 14ns x 64 x 10 = 8960ns
    FM3_ADC2->ADSS2 = (0 << 1);
    FM3_ADC2->CMPCR = 17;
    FM3_ADC2->ADSR = ADSR_FDAS;        // ADSTP(R/W):0, FDAS(R/W):1(LSB), , , , PCNS(R):0, PCS(R):0, SCS(R):0
    FM3_ADC2->ADCR = 0;                // SCIF:0, PCIF:0, CMPIF:0, , SCIE:0 , PCIE:0 , CMPIE:0, OVRIE:0
    FM3_ADC2->SCCR |= SCCR_SFCLR;
    FM3_ADC2->SCCR &= ~SCCR_RPT;
    g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inX = 0;
    g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inY = 0;
    g_CQ_FRK_FM3_TouchPanel_Driver.m_global_touch = 0;
    g_CQ_FRK_FM3_TouchPanel_Driver.m_samplingTimespan = FM3_TOUCHPANEL_POLL_TIMESPAN;
    g_CQ_FRK_FM3_TouchPanel_Driver.m_touchCompletion.Initialize();
    g_CQ_FRK_FM3_TouchPanel_Driver.m_touchCompletion.InitializeForISR(CQ_FRK_FM3_TouchPanel_Driver::TouchCompletion, NULL);
    g_CQ_FRK_FM3_TouchPanel_Driver.m_touchCompletion.EnqueueDelta(g_CQ_FRK_FM3_TouchPanel_Driver.m_samplingTimespan);
    if (!::CPU_GPIO_EnableInputPin2(
            0xB4,
            FALSE,
            touchIsrProc,
            NULL,
            GPIO_INT_EDGE_BOTH,
            RESISTOR_PULLUP))
    {
        return FALSE;
    }
    return TRUE;
}

BOOL CQ_FRK_FM3_TouchPanel_Driver::Uninitialize()
{
    if (g_CQ_FRK_FM3_TouchPanel_Driver.m_touchCompletion.IsLinked())
    {
        g_CQ_FRK_FM3_TouchPanel_Driver.m_touchCompletion.Abort();
    }
    return TRUE;
}

void CQ_FRK_FM3_TouchPanel_Driver::Poll()
{
    static UINT32 t_state;
    UINT32 value;

    switch(t_state) {
    case 0:
        SENSE_ON;
        t_state++;
        break;
    case 1:
        if (TOUCH_CHK)
            t_state++;
        else {
            IDLE;
            g_CQ_FRK_FM3_TouchPanel_Driver.m_global_touch = 0;
            t_state = 0;
        }
        break;
    case 2:    // skip for chattering
    case 3:
        t_state++;
        break;
#if 0
    case 4:
        if (TOUCH_CHK) {
            SENSE_Y_ON;
            t_state++;
        } else {
            IDLE;
            g_CQ_FRK_FM3_TouchPanel_Driver.m_global_touch = 0;
            t_state = 0;
        }
        break;
    case 5: // start A-D for getting Y axis
        AD_Y_ENABLE;
        FM3_ADC1->SCCR |= SCCR_SSTR;    // A/D start
        t_state++;
        break;
    case 6: // sense power on
        value = FM3_ADC1->SCFD;
        g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inX = (UINT16)(value >> 16);
        AD_Y_DISABLE;
        SENSE_ON;
        t_state++;
        break;
#else
    case 4:
        if (TOUCH_CHK) {
            GLOBAL_LOCK(irq);
            SENSE_Y_ON;
            AD_Y_ENABLE;
            volatile int i = 10000;
            while (i-- > 0);
            FM3_ADC1->SCCR |= SCCR_SSTR;    // A/D start
            do {
                value = FM3_ADC1->SCFD;
            } while (((value & SCFDL_INVL) != 0) || (16 != (value & 0x1F)));
            g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inY = (UINT16)(value >> 16);
            AD_Y_DISABLE;
            SENSE_ON;
            t_state += 3;
        } else {
            IDLE;
            g_CQ_FRK_FM3_TouchPanel_Driver.m_global_touch = 0;
            t_state = 0;
        }
        break;
#endif
    case 7:
        if (TOUCH_CHK) {
            SENSE_X_ON;
            t_state++;
        } else {
            IDLE;
            g_CQ_FRK_FM3_TouchPanel_Driver.m_global_touch = 0;
            t_state = 0;
        }
        break;
    case 8:    // start A-D for sensing X axis
        AD_X_ENABLE;
        FM3_ADC2->SCCR |= SCCR_SSTR;    // A/D start
        t_state++;
        break;
    case 9:    // sense power on
        value = FM3_ADC2->SCFD;
        g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inX = (UINT16)(value >> 16);
        AD_X_DISABLE;
        SENSE_ON;
        t_state++;
        break;
    case 10:    //  skip for chattering
    case 11:
        t_state++;
        break;
    case 12:
        if (TOUCH_CHK) {
            g_CQ_FRK_FM3_TouchPanel_Driver.m_global_touch = 1;
#ifdef DEBUG_TOUCHPANEL
            lcd_printf("x=%d y=%d\r\n",
                    g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inX,
                    g_CQ_FRK_FM3_TouchPanel_Driver.m_global_inY);
#endif
        }
        IDLE;
        t_state = 0;
        break;
    default:
        t_state = 0;
        break;
    }
}
